from concurrent.futures import ThreadPoolExecutor
import os
import sys
import time

from metaflow.exception import MetaflowException
from metaflow.metaflow_config import KUBERNETES_NAMESPACE

from .kubernetes_job import KubernetesJob, KubernetesJobSet

CLIENT_REFRESH_INTERVAL_SECONDS = 300


class KubernetesClientException(MetaflowException):
    headline = "Kubernetes client error"


class KubernetesClient(object):
    def __init__(self):
        try:
            # Kubernetes is a soft dependency.
            from kubernetes import client, config
        except (NameError, ImportError):
            raise KubernetesClientException(
                "Could not import module 'kubernetes'.\n\nInstall Kubernetes "
                "Python package (https://pypi.org/project/kubernetes/) first.\n"
                "You can install the module by executing - "
                "%s -m pip install kubernetes\n"
                "or equivalent through your favorite Python package manager."
                % sys.executable
            )
        self._refresh_client()
        self._namespace = KUBERNETES_NAMESPACE

    def _refresh_client(self):
        from kubernetes import client, config

        if os.getenv("KUBECONFIG"):
            # There are cases where we're running inside a pod, but can't use
            # the kubernetes client for that pod's cluster: for example when
            # running in Bitbucket Cloud or other CI system.
            # In this scenario, the user can set a KUBECONFIG environment variable
            # to load the kubeconfig, regardless of whether we're in a pod or not.
            config.load_kube_config()
        elif os.getenv("KUBERNETES_SERVICE_HOST"):
            # We are inside a pod, authenticate via ServiceAccount assigned to us
            config.load_incluster_config()
        else:
            # Default to using kubeconfig, likely $HOME/.kube/config
            # TODO (savin):
            #  1. Support generating kubeconfig on the fly using boto3
            #  2. Support auth via OIDC - https://docs.aws.amazon.com/eks/latest/userguide/authenticate-oidc-identity-provider.html
            config.load_kube_config()
        self._client = client
        self._client_refresh_timestamp = time.time()

    def get(self):
        if (
            time.time() - self._client_refresh_timestamp
            > CLIENT_REFRESH_INTERVAL_SECONDS
        ):
            self._refresh_client()

        return self._client

    def _find_active_pods(self, flow_name, run_id=None, user=None):
        def _request(_continue=None):
            # handle paginated responses
            return self._client.CoreV1Api().list_namespaced_pod(
                namespace=self._namespace,
                # limited selector support for K8S api. We want to cover multiple statuses: Running / Pending / Unknown
                field_selector="status.phase!=Succeeded,status.phase!=Failed",
                limit=1000,
                _continue=_continue,
            )

        results = _request()

        if run_id is not None:
            # handle argo prefixes in run_id
            run_id = run_id[run_id.startswith("argo-") and len("argo-") :]

        while results.metadata._continue or results.items:
            for pod in results.items:
                match = (
                    # arbitrary pods might have no annotations at all.
                    pod.metadata.annotations
                    and pod.metadata.labels
                    and (
                        run_id is None
                        or (pod.metadata.annotations.get("metaflow/run_id") == run_id)
                        # we want to also match pods launched by argo-workflows
                        or (
                            pod.metadata.labels.get("workflows.argoproj.io/workflow")
                            == run_id
                        )
                    )
                    and (
                        user is None
                        or pod.metadata.annotations.get("metaflow/user") == user
                    )
                    and (
                        pod.metadata.annotations.get("metaflow/flow_name") == flow_name
                    )
                )
                if match:
                    yield pod
            if not results.metadata._continue:
                break
            results = _request(results.metadata._continue)

    def list(self, flow_name, run_id, user):
        results = self._find_active_pods(flow_name, run_id, user)

        return list(results)

    def kill_pods(self, flow_name, run_id, user, echo):
        from kubernetes.stream import stream

        api_instance = self._client.CoreV1Api()
        job_api = self._client.BatchV1Api()
        pods = self._find_active_pods(flow_name, run_id, user)

        def _kill_pod(pod):
            echo("Killing Kubernetes pod %s\n" % pod.metadata.name)
            try:
                stream(
                    api_instance.connect_get_namespaced_pod_exec,
                    name=pod.metadata.name,
                    namespace=pod.metadata.namespace,
                    command=[
                        "/bin/sh",
                        "-c",
                        "/sbin/killall5",
                    ],
                    stderr=True,
                    stdin=False,
                    stdout=True,
                    tty=False,
                )
            except Exception:
                # best effort kill for pod can fail.
                try:
                    job_name = pod.metadata.labels.get("job-name", None)
                    if job_name is None:
                        raise Exception("Could not determine job name")

                    job_api.patch_namespaced_job(
                        name=job_name,
                        namespace=pod.metadata.namespace,
                        field_manager="metaflow",
                        body={"spec": {"parallelism": 0}},
                    )
                except Exception as e:
                    echo("failed to kill pod %s - %s" % (pod.metadata.name, str(e)))

        with ThreadPoolExecutor() as executor:
            operated_pods = list(executor.map(_kill_pod, pods))

            if not operated_pods:
                echo("No active Kubernetes pods found for run *%s*" % run_id)

    def jobset(self, **kwargs):
        return KubernetesJobSet(self, **kwargs)

    def job(self, **kwargs):
        return KubernetesJob(self, **kwargs)



import tarfile

from io import BytesIO
from typing import IO, List, Optional, Union

from .backend import PackagingBackend


class TarPackagingBackend(PackagingBackend):
    type = "tgz"

    @classmethod
    def get_extract_commands(cls, archive_name: str, dest_dir: str) -> List[str]:
        return [
            f"TAR_OPTIONS='--warning=no-timestamp' tar -xzf {archive_name} -C {dest_dir}"
        ]

    def __init__(self):
        super().__init__()
        self._buf = None

    def create(self):
        self._buf = BytesIO()
        self._archive = tarfile.open(
            fileobj=self._buf, mode="w:gz", compresslevel=3, dereference=True
        )
        return self

    def add_file(self, filename: str, arcname: Optional[str] = None):
        info = self._archive.gettarinfo(filename, arcname)
        # Setting this default to Dec 3, 2019
        info.mtime = 1575360000
        with open(filename, mode="rb") as f:
            self._archive.addfile(info, f)

    def add_data(self, data: BytesIO, arcname: str):
        info = tarfile.TarInfo(arcname)
        data.seek(0)
        info.size = len(data.getvalue())
        # Setting this default to Dec 3, 2019
        info.mtime = 1575360000
        self._archive.addfile(info, data)

    def close(self):
        if self._archive:
            self._archive.close()

    def get_blob(self) -> Optional[Union[bytes, bytearray]]:
        if self._buf:
            blob = bytearray(self._buf.getvalue())
            blob[4:8] = [0] * 4  # Reset 4 bytes from offset 4 to account for ts
            return blob
        return None

    @classmethod
    def cls_open(cls, content: IO[bytes]) -> tarfile.TarFile:
        return tarfile.open(fileobj=content, mode="r:gz")

    @classmethod
    def cls_has_member(cls, archive: tarfile.TarFile, name: str) -> bool:
        try:
            archive.getmember(name)
            return True
        except KeyError:
            return False

    @classmethod
    def cls_get_member(cls, archive: tarfile.TarFile, name: str) -> Optional[bytes]:
        try:
            member = archive.getmember(name)
            return archive.extractfile(member).read()
        except KeyError:
            return None

    @classmethod
    def cls_extract_members(
        cls,
        archive: tarfile.TarFile,
        members: Optional[List[str]] = None,
        dest_dir: str = ".",
    ) -> None:
        archive.extractall(path=dest_dir, members=members)

    @classmethod
    def cls_list_members(cls, archive: tarfile.TarFile) -> Optional[List[str]]:
        return archive.getnames() or None



import os

from metaflow.exception import MetaflowException
from metaflow.metaflow_environment import MetaflowEnvironment
from metaflow.packaging_sys import ContentType


class UVException(MetaflowException):
    headline = "uv error"


class UVEnvironment(MetaflowEnvironment):
    TYPE = "uv"

    def __init__(self, flow):
        super().__init__(flow)
        self.flow = flow

    def validate_environment(self, logger, datastore_type):
        self.datastore_type = datastore_type
        self.logger = logger

    def init_environment(self, echo, only_steps=None):
        self.logger("Bootstrapping uv...")

    def executable(self, step_name, default=None):
        return "uv run --no-sync python"

    def add_to_package(self):
        # NOTE: We treat uv.lock and pyproject.toml as regular project assets and ship these along user code as part of the code package
        # These are the minimal required files to reproduce the UV environment on the remote platform.
        def _find(filename):
            current_dir = os.getcwd()
            while True:
                file_path = os.path.join(current_dir, filename)
                if os.path.isfile(file_path):
                    return file_path
                parent_dir = os.path.dirname(current_dir)
                if parent_dir == current_dir:  # Reached root
                    raise UVException(
                        f"Could not find {filename} in current directory or any parent directory"
                    )
                current_dir = parent_dir

        pyproject_path = _find("pyproject.toml")
        uv_lock_path = _find("uv.lock")
        files = [
            (uv_lock_path, "uv.lock", ContentType.OTHER_CONTENT),
            (pyproject_path, "pyproject.toml", ContentType.OTHER_CONTENT),
        ]
        return files

    def pylint_config(self):
        config = super().pylint_config()
        # Disable (import-error) in pylint
        config.append("--disable=F0401")
        return config

    def bootstrap_commands(self, step_name, datastore_type):
        return [
            "echo 'Bootstrapping uv project...'",
            "flush_mflogs",
            # We have to prevent the tracing module from loading, as the bootstrapping process
            # uses the internal S3 client which would fail to import tracing due to the required
            # dependencies being bundled into the conda environment, which is yet to be
            # initialized at this point.
            'DISABLE_TRACING=True python -m metaflow.plugins.uv.bootstrap "%s"'
            % datastore_type,
            "echo 'uv project bootstrapped.'",
            "flush_mflogs",
            "export PATH=$PATH:$(pwd)/uv_install",
        ]



import sys
import json
import time
import tempfile
from typing import ClassVar, Optional

from metaflow.client.core import get_metadata
from metaflow.exception import MetaflowException
from metaflow.plugins.argo.argo_client import ArgoClient
from metaflow.metaflow_config import KUBERNETES_NAMESPACE
from metaflow.plugins.argo.argo_workflows import ArgoWorkflows
from metaflow.runner.deployer import (
    Deployer,
    DeployedFlow,
    TriggeredRun,
    generate_fake_flow_file_contents,
)

from metaflow.runner.utils import get_lower_level_group, handle_timeout, temporary_fifo


class ArgoWorkflowsTriggeredRun(TriggeredRun):
    """
    A class representing a triggered Argo Workflow execution.
    """

    def suspend(self, **kwargs) -> bool:
        """
        Suspend the running workflow.

        Parameters
        ----------
        authorize : str, optional, default None
            Authorize the suspension with a production token.

        Returns
        -------
        bool
            True if the command was successful, False otherwise.
        """
        _, run_id = self.pathspec.split("/")

        # every subclass needs to have `self.deployer_kwargs`
        command = get_lower_level_group(
            self.deployer.api,
            self.deployer.top_level_kwargs,
            self.deployer.TYPE,
            self.deployer.deployer_kwargs,
        ).suspend(run_id=run_id, **kwargs)

        pid = self.deployer.spm.run_command(
            [sys.executable, *command],
            env=self.deployer.env_vars,
            cwd=self.deployer.cwd,
            show_output=self.deployer.show_output,
        )

        command_obj = self.deployer.spm.get(pid)
        command_obj.sync_wait()
        return command_obj.process.returncode == 0

    def unsuspend(self, **kwargs) -> bool:
        """
        Unsuspend the suspended workflow.

        Parameters
        ----------
        authorize : str, optional, default None
            Authorize the unsuspend with a production token.

        Returns
        -------
        bool
            True if the command was successful, False otherwise.
        """
        _, run_id = self.pathspec.split("/")

        # every subclass needs to have `self.deployer_kwargs`
        command = get_lower_level_group(
            self.deployer.api,
            self.deployer.top_level_kwargs,
            self.deployer.TYPE,
            self.deployer.deployer_kwargs,
        ).unsuspend(run_id=run_id, **kwargs)

        pid = self.deployer.spm.run_command(
            [sys.executable, *command],
            env=self.deployer.env_vars,
            cwd=self.deployer.cwd,
            show_output=self.deployer.show_output,
        )

        command_obj = self.deployer.spm.get(pid)
        command_obj.sync_wait()
        return command_obj.process.returncode == 0

    def terminate(self, **kwargs) -> bool:
        """
        Terminate the running workflow.

        Parameters
        ----------
        authorize : str, optional, default None
            Authorize the termination with a production token.

        Returns
        -------
        bool
            True if the command was successful, False otherwise.
        """
        _, run_id = self.pathspec.split("/")

        # every subclass needs to have `self.deployer_kwargs`
        command = get_lower_level_group(
            self.deployer.api,
            self.deployer.top_level_kwargs,
            self.deployer.TYPE,
            self.deployer.deployer_kwargs,
        ).terminate(run_id=run_id, **kwargs)

        pid = self.deployer.spm.run_command(
            [sys.executable, *command],
            env=self.deployer.env_vars,
            cwd=self.deployer.cwd,
            show_output=self.deployer.show_output,
        )

        command_obj = self.deployer.spm.get(pid)
        command_obj.sync_wait()
        return command_obj.process.returncode == 0

    def wait_for_completion(
        self, check_interval: int = 5, timeout: Optional[int] = None
    ):
        """
        Wait for the workflow to complete or timeout.

        Parameters
        ----------
        check_interval: int, default: 5
            Frequency of checking for workflow completion, in seconds.
        timeout : int, optional, default None
            Maximum time in seconds to wait for workflow completion.
            If None, waits indefinitely.

        Raises
        ------
        TimeoutError
            If the workflow does not complete within the specified timeout period.
        """
        start_time = time.time()
        while self.is_running:
            if timeout is not None and (time.time() - start_time) > timeout:
                raise TimeoutError(
                    "Workflow did not complete within specified timeout."
                )
            time.sleep(check_interval)

    @property
    def is_running(self):
        """
        Check if the workflow is currently running.

        Returns
        -------
        bool
            True if the workflow status is either 'Pending' or 'Running',
            False otherwise.
        """
        workflow_status = self.status
        # full list of all states present here:
        # https://github.com/argoproj/argo-workflows/blob/main/pkg/apis/workflow/v1alpha1/workflow_types.go#L54
        # we only consider non-terminal states to determine if the workflow has not finished
        return workflow_status is not None and workflow_status in ["Pending", "Running"]

    @property
    def status(self) -> Optional[str]:
        """
        Get the status of the triggered run.

        Returns
        -------
        str, optional
            The status of the workflow considering the run object, or None if
            the status could not be retrieved.
        """
        from metaflow.plugins.argo.argo_workflows_cli import (
            get_status_considering_run_object,
        )

        flow_name, run_id = self.pathspec.split("/")
        name = run_id[5:]
        status = ArgoWorkflows.get_workflow_status(flow_name, name)
        if status is not None:
            return get_status_considering_run_object(status, self.run)
        return None


class ArgoWorkflowsDeployedFlow(DeployedFlow):
    """
    A class representing a deployed Argo Workflow template.
    """

    TYPE: ClassVar[Optional[str]] = "argo-workflows"

    @classmethod
    def list_deployed_flows(cls, flow_name: Optional[str] = None):
        """
        List all deployed Argo Workflow templates.

        Parameters
        ----------
        flow_name : str, optional, default None
            If specified, only list deployed flows for this specific flow name.
            If None, list all deployed flows.

        Yields
        ------
        ArgoWorkflowsDeployedFlow
            `ArgoWorkflowsDeployedFlow` objects representing deployed
            workflow templates on Argo Workflows.
        """
        from metaflow.plugins.argo.argo_workflows import ArgoWorkflows

        # When flow_name is None, use all=True to get all templates
        # When flow_name is specified, use all=False to filter by flow_name
        all_templates = flow_name is None
        for template_name in ArgoWorkflows.list_templates(
            flow_name=flow_name, all=all_templates
        ):
            try:
                deployed_flow = cls.from_deployment(template_name)
                yield deployed_flow
            except Exception:
                # Skip templates that can't be converted to DeployedFlow objects
                continue

    @classmethod
    def from_deployment(cls, identifier: str, metadata: Optional[str] = None):
        """
        Retrieves a `ArgoWorkflowsDeployedFlow` object from an identifier and optional
        metadata.

        Parameters
        ----------
        identifier : str
            Deployer specific identifier for the workflow to retrieve
        metadata : str, optional, default None
            Optional deployer specific metadata.

        Returns
        -------
        ArgoWorkflowsDeployedFlow
            A `ArgoWorkflowsDeployedFlow` object representing the
            deployed flow on argo workflows.
        """
        client = ArgoClient(namespace=KUBERNETES_NAMESPACE)
        workflow_template = client.get_workflow_template(identifier)

        if workflow_template is None:
            raise MetaflowException("No deployed flow found for: %s" % identifier)

        metadata_annotations = workflow_template.get("metadata", {}).get(
            "annotations", {}
        )

        flow_name = metadata_annotations.get("metaflow/flow_name", "")
        username = metadata_annotations.get("metaflow/owner", "")
        parameters = json.loads(metadata_annotations.get("metaflow/parameters", "{}"))

        # these two only exist if @project decorator is used..
        branch_name = metadata_annotations.get("metaflow/branch_name", None)
        project_name = metadata_annotations.get("metaflow/project_name", None)

        project_kwargs = {}
        if branch_name is not None:
            if branch_name.startswith("prod."):
                project_kwargs["production"] = True
                project_kwargs["branch"] = branch_name[len("prod.") :]
            elif branch_name.startswith("test."):
                project_kwargs["branch"] = branch_name[len("test.") :]
            elif branch_name == "prod":
                project_kwargs["production"] = True

        fake_flow_file_contents = generate_fake_flow_file_contents(
            flow_name=flow_name, param_info=parameters, project_name=project_name
        )

        with tempfile.NamedTemporaryFile(suffix=".py", delete=False) as fake_flow_file:
            with open(fake_flow_file.name, "w") as fp:
                fp.write(fake_flow_file_contents)

            if branch_name is not None:
                d = Deployer(
                    fake_flow_file.name,
                    env={"METAFLOW_USER": username},
                    **project_kwargs,
                ).argo_workflows()
            else:
                d = Deployer(
                    fake_flow_file.name, env={"METAFLOW_USER": username}
                ).argo_workflows(name=identifier)

            d.name = identifier
            d.flow_name = flow_name
            if metadata is None:
                d.metadata = get_metadata()
            else:
                d.metadata = metadata

        return cls(deployer=d)

    @property
    def production_token(self) -> Optional[str]:
        """
        Get the production token for the deployed flow.

        Returns
        -------
        str, optional
            The production token, None if it cannot be retrieved.
        """
        try:
            _, production_token = ArgoWorkflows.get_existing_deployment(
                self.deployer.name
            )
            return production_token
        except TypeError:
            return None

    def delete(self, **kwargs) -> bool:
        """
        Delete the deployed workflow template.

        Parameters
        ----------
        authorize : str, optional, default None
            Authorize the deletion with a production token.

        Returns
        -------
        bool
            True if the command was successful, False otherwise.
        """
        command = get_lower_level_group(
            self.deployer.api,
            self.deployer.top_level_kwargs,
            self.deployer.TYPE,
            self.deployer.deployer_kwargs,
        ).delete(**kwargs)

        pid = self.deployer.spm.run_command(
            [sys.executable, *command],
            env=self.deployer.env_vars,
            cwd=self.deployer.cwd,
            show_output=self.deployer.show_output,
        )

        command_obj = self.deployer.spm.get(pid)
        command_obj.sync_wait()
        return command_obj.process.returncode == 0

    def trigger(self, **kwargs) -> ArgoWorkflowsTriggeredRun:
        """
        Trigger a new run for the deployed flow.

        Parameters
        ----------
        **kwargs : Any
            Additional arguments to pass to the trigger command,
            `Parameters` in particular.

        Returns
        -------
        ArgoWorkflowsTriggeredRun
            The triggered run instance.

        Raises
        ------
        Exception
            If there is an error during the trigger process.
        """
        with temporary_fifo() as (attribute_file_path, attribute_file_fd):
            # every subclass needs to have `self.deployer_kwargs`
            command = get_lower_level_group(
                self.deployer.api,
                self.deployer.top_level_kwargs,
                self.deployer.TYPE,
                self.deployer.deployer_kwargs,
            ).trigger(deployer_attribute_file=attribute_file_path, **kwargs)

            pid = self.deployer.spm.run_command(
                [sys.executable, *command],
                env=self.deployer.env_vars,
                cwd=self.deployer.cwd,
                show_output=self.deployer.show_output,
            )

            command_obj = self.deployer.spm.get(pid)
            content = handle_timeout(
                attribute_file_fd, command_obj, self.deployer.file_read_timeout
            )
            command_obj.sync_wait()
            if command_obj.process.returncode == 0:
                return ArgoWorkflowsTriggeredRun(
                    deployer=self.deployer, content=content
                )

        raise Exception(
            "Error triggering %s on %s for %s"
            % (
                self.deployer.name,
                self.deployer.TYPE,
                self.deployer.flow_file,
            )
        )



from functools import partial
from typing import Any, Dict, Generator, List, Optional, Tuple, TYPE_CHECKING, Union

from metaflow.debug import debug
from metaflow.exception import MetaflowException
from metaflow.user_configs.config_parameters import ConfigValue

if TYPE_CHECKING:
    import metaflow.flowspec
    import metaflow.parameters
    import metaflow.user_decorators.mutable_step


class MutableFlow:
    IGNORE = 1
    ERROR = 2
    OVERRIDE = 3

    def __init__(
        self,
        flow_spec: "metaflow.flowspec.FlowSpec",
        pre_mutate: bool = False,
        statically_defined: bool = False,
        inserted_by: Optional[str] = None,
    ):
        self._flow_cls = flow_spec
        self._pre_mutate = pre_mutate
        self._statically_defined = statically_defined
        self._inserted_by = inserted_by
        if self._inserted_by is None:
            # This is an error because MutableSteps should only be created by
            # StepMutators or FlowMutators. We need to catch it now because otherwise
            # we may put stuff on the command line (with --with) that would get added
            # twice and weird behavior may ensue.
            raise MetaflowException(
                "MutableFlow should only be created by StepMutators or FlowMutators. "
                "This is an internal error."
            )

    @property
    def decorator_specs(
        self,
    ) -> Generator[Tuple[str, str, List[Any], Dict[str, Any]], None, None]:
        """
        Iterate over all the decorator specifications of this flow. Note that the same
        type of decorator may be present multiple times and no order is guaranteed.

        The returned tuple contains:
        - The decorator's name (shortest possible)
        - The decorator's fully qualified name (in the case of Metaflow decorators, this
          will indicate which extension the decorator comes from)
        - A list of positional arguments
        - A dictionary of keyword arguments

        You can use the decorator specification to remove a decorator from the flow
        for example.

        Yields
        ------
        str, str, List[Any], Dict[str, Any]
            A tuple containing the decorator name, it's fully qualified name,
            a list of positional arguments, and a dictionary of keyword arguments.
        """
        for decos in self._flow_cls._flow_decorators.values():
            for deco in decos:
                # 3.7 does not support yield foo, *bar syntax so we
                # work around

                r = [
                    deco.name,
                    "%s.%s"
                    % (
                        deco.__class__.__module__,
                        deco.__class__.__name__,
                    ),
                ]
                r.extend(deco.get_args_kwargs())
                yield tuple(r)

    @property
    def configs(self) -> Generator[Tuple[str, ConfigValue], None, None]:
        """
        Iterate over all user configurations in this flow

        Use this to parameterize your flow based on configuration. As an example, the
        `pre_mutate`/`mutate` methods can add decorators to steps in the flow that
        depend on values in the configuration.

        ```
        class MyDecorator(FlowMutator):
            def mutate(flow: MutableFlow):
                val = next(flow.configs)[1].steps.start.cpu
                flow.start.add_decorator(environment, vars={'mycpu': val})
                return flow

        @MyDecorator()
        class TestFlow(FlowSpec):
            config = Config('myconfig.json')

            @step
            def start(self):
                pass
        ```
        can be used to add an environment decorator to the `start` step.

        Yields
        ------
        Tuple[str, ConfigValue]
            Iterates over the configurations of the flow
        """
        from metaflow.flowspec import _FlowState

        # When configs are parsed, they are loaded in _flow_state[_FlowState.CONFIGS]
        for name, value in self._flow_cls._flow_state.get(
            _FlowState.CONFIGS, {}
        ).items():
            r = name, ConfigValue(value)
            debug.userconf_exec("Mutable flow yielding config: %s" % str(r))
            yield r

    @property
    def parameters(
        self,
    ) -> Generator[Tuple[str, "metaflow.parameters.Parameter"], None, None]:
        """
        Iterate over all the parameters in this flow.

        Yields
        ------
        Tuple[str, Parameter]
            Name of the parameter and parameter in the flow
        """
        for var, param in self._flow_cls._get_parameters():
            if param.IS_CONFIG_PARAMETER:
                continue
            debug.userconf_exec(
                "Mutable flow yielding parameter: %s" % str((var, param))
            )
            yield var, param

    @property
    def steps(
        self,
    ) -> Generator[
        Tuple[str, "metaflow.user_decorators.mutable_step.MutableStep"], None, None
    ]:
        """
        Iterate over all the steps in this flow. The order of the steps
        returned is not guaranteed.

        Yields
        ------
        Tuple[str, MutableStep]
            A tuple with the step name and the step proxy
        """
        from .mutable_step import MutableStep

        for var in dir(self._flow_cls):
            potential_step = getattr(self._flow_cls, var)
            if callable(potential_step) and hasattr(potential_step, "is_step"):
                debug.userconf_exec("Mutable flow yielding step: %s" % var)
                yield var, MutableStep(
                    self._flow_cls,
                    potential_step,
                    pre_mutate=self._pre_mutate,
                    statically_defined=self._statically_defined,
                    inserted_by=self._inserted_by,
                )

    def add_parameter(
        self, name: str, value: "metaflow.parameters.Parameter", overwrite: bool = False
    ) -> None:
        """
        Add a parameter to the flow. You can only add parameters in the `pre_mutate`
        method.

        Parameters
        ----------
        name : str
            Name of the parameter
        value : Parameter
            Parameter to add to the flow
        overwrite : bool, default False
            If True, overwrite the parameter if it already exists
        """
        if not self._pre_mutate:
            raise MetaflowException(
                "Adding parameter '%s' from %s is only allowed in the `pre_mutate` "
                "method and not the `mutate` method" % (name, self._inserted_by)
            )
        from metaflow.parameters import Parameter

        if hasattr(self._flow_cls, name) and not overwrite:
            raise MetaflowException(
                "Flow '%s' already has a class member '%s' -- "
                "set overwrite=True in add_parameter to overwrite it."
                % (self._flow_cls.__name__, name)
            )
        if not isinstance(value, Parameter) or value.IS_CONFIG_PARAMETER:
            raise MetaflowException(
                "Only a Parameter or an IncludeFile can be added using `add_parameter`"
                "; got %s" % type(value)
            )
        debug.userconf_exec("Mutable flow adding parameter %s to flow" % name)
        setattr(self._flow_cls, name, value)

    def remove_parameter(self, parameter_name: str) -> bool:
        """
        Remove a parameter from the flow.

        The name given should match the name of the parameter (can be different
        from the name of the parameter in the flow. You can not remove config parameters.
        You can only remove parameters in the `pre_mutate` method.

        Parameters
        ----------
        parameter_name : str
            Name of the parameter

        Returns
        -------
        bool
            Returns True if the parameter was removed
        """
        if not self._pre_mutate:
            raise MetaflowException(
                "Removing parameter '%s' from %s is only allowed in the `pre_mutate` "
                "method and not the `mutate` method"
                % (parameter_name, " from ".join(self._inserted_by))
            )
        from metaflow.flowspec import _FlowState

        for var, param in self._flow_cls._get_parameters():
            if param.IS_CONFIG_PARAMETER:
                continue
            if param.name == parameter_name:
                delattr(self._flow_cls, var)
                debug.userconf_exec(
                    "Mutable flow removing parameter %s from flow" % var
                )
                # Reset so that we don't list it again
                del self._flow_cls._flow_state[_FlowState.CACHED_PARAMETERS]
                return True
        debug.userconf_exec(
            "Mutable flow failed to remove parameter %s from flow" % parameter_name
        )
        return False

    def add_decorator(
        self,
        deco_type: Union[partial, str],
        deco_args: Optional[List[Any]] = None,
        deco_kwargs: Optional[Dict[str, Any]] = None,
        duplicates: int = IGNORE,
    ) -> None:
        """
        Add a Metaflow flow-decorator to a flow. You can only add decorators in the
        `pre_mutate` method.

        You can either add the decorator itself or its decorator specification for it
        (the same you would get back from decorator_specs). You can also mix and match
        but you cannot provide arguments both through the string and the
        deco_args/deco_kwargs.

        As an example:
        ```
        from metaflow import project

        ...
        my_flow.add_decorator(project, deco_kwargs={"name":"my_project"})
        ```

        is equivalent to:
        ```
        my_flow.add_decorator("project:name=my_project")
        ```

        Note in the later case, there is no need to import the flow decorator.

        The latter syntax is useful to, for example, allow decorators to be stored as
        strings in a configuration file.

        In terms of precedence for decorators:
          - if a decorator can be applied multiple times all decorators
            added are kept (this is rare for flow-decorators).
          - if `duplicates` is set to `MutableFlow.IGNORE`, then the decorator
            being added is ignored (in other words, the existing decorator has precedence).
          - if `duplicates` is set to `MutableFlow.OVERRIDE`, then the *existing*
            decorator is removed and this newly added one replaces it (in other
            words, the newly added decorator has precedence).
          - if `duplicates` is set to `MutableFlow.ERROR`, then an error is raised but only
            if the newly added decorator is *static* (ie: defined directly in the code).
            If not, it is ignored.

        Parameters
        ----------
        deco_type : Union[partial, str]
            The decorator class to add to this flow. If using a string, you cannot specify
            additional arguments as all argument will be parsed from the decorator
            specification.
        deco_args : List[Any], optional, default None
            Positional arguments to pass to the decorator.
        deco_kwargs : Dict[str, Any], optional, default None
            Keyword arguments to pass to the decorator.
        duplicates : int, default MutableFlow.IGNORE
            Instruction on how to handle duplicates. It can be one of:
            - `MutableFlow.IGNORE`: Ignore the decorator if it already exists.
            - `MutableFlow.ERROR`: Raise an error if the decorator already exists.
            - `MutableFlow.OVERRIDE`: Remove the existing decorator and add this one.

        """
        if not self._pre_mutate:
            raise MetaflowException(
                "Adding flow-decorator '%s' from %s is only allowed in the `pre_mutate` "
                "method and not the `mutate` method"
                % (
                    deco_type if isinstance(deco_type, str) else deco_type.name,
                    self._inserted_by,
                )
            )
        # Prevent circular import
        from metaflow.decorators import (
            DuplicateFlowDecoratorException,
            FlowDecorator,
            extract_flow_decorator_from_decospec,
        )

        deco_args = deco_args or []
        deco_kwargs = deco_kwargs or {}

        def _add_flow_decorator(flow_deco):
            if deco_args:
                raise MetaflowException(
                    "Flow decorators do not take additional positional arguments"
                )
            # Update kwargs:
            flow_deco.attributes.update(deco_kwargs)

            # Check duplicates
            def _do_add():
                flow_deco.statically_defined = self._statically_defined
                flow_deco.inserted_by = self._inserted_by
                self._flow_cls._flow_decorators.setdefault(flow_deco.name, []).append(
                    flow_deco
                )
                debug.userconf_exec(
                    "Mutable flow adding flow decorator '%s'" % deco_type
                )

            # self._flow_cls._flow_decorators is a  dictionary of form :
            # <deco_name> : [deco_instance, deco_instance, ...]
            existing_deco = [
                d for d in self._flow_cls._flow_decorators if d == flow_deco.name
            ]

            if flow_deco.allow_multiple or not existing_deco:
                _do_add()
            elif duplicates == MutableFlow.IGNORE:
                # If we ignore, we do not add the decorator
                debug.userconf_exec(
                    "Mutable flow ignoring flow decorator '%s'"
                    "(already exists and duplicates are ignored)" % flow_deco.name
                )
            elif duplicates == MutableFlow.OVERRIDE:
                # If we override, we remove the existing decorator and add this one
                debug.userconf_exec(
                    "Mutable flow overriding flow decorator '%s' "
                    "(removing existing decorator and adding new one)" % flow_deco.name
                )
                self._flow_cls._flow_decorators = [
                    d
                    for d in self._flow_cls._flow_decorators
                    if d.name != flow_deco.name
                ]
                _do_add()
            elif duplicates == MutableFlow.ERROR:
                # If we error, we raise an exception
                if self._statically_defined:
                    raise DuplicateFlowDecoratorException(flow_deco.name)
                else:
                    debug.userconf_exec(
                        "Mutable flow ignoring flow decorator '%s' "
                        "(already exists and non statically defined)" % flow_deco.name
                    )
            else:
                raise ValueError("Invalid duplicates value: %s" % duplicates)

        # If deco_type is a string, we want to parse it to a decospec
        if isinstance(deco_type, str):
            flow_deco, has_args_kwargs = extract_flow_decorator_from_decospec(deco_type)
            if (deco_args or deco_kwargs) and has_args_kwargs:
                raise MetaflowException(
                    "Cannot specify additional arguments when adding a flow decorator "
                    "using a decospec that already contains arguments"
                )
            _add_flow_decorator(flow_deco)
            return

        # Validate deco_type
        if (
            not isinstance(deco_type, partial)
            or len(deco_type.args) != 1
            or not issubclass(deco_type.args[0], FlowDecorator)
        ):
            raise TypeError("add_decorator takes a FlowDecorator")

        deco_type = deco_type.args[0]
        _add_flow_decorator(
            deco_type(
                attributes=deco_kwargs,
                statically_defined=self._statically_defined,
                inserted_by=self._inserted_by,
            )
        )

    def remove_decorator(
        self,
        deco_name: str,
        deco_args: Optional[List[Any]] = None,
        deco_kwargs: Optional[Dict[str, Any]] = None,
    ) -> bool:
        """
        Remove a flow-level decorator. To remove a decorator, you can pass the decorator
        specification (obtained from `decorator_specs` for example).
        Note that if multiple decorators share the same decorator specification
        (very rare), they will all be removed.

        You can only remove decorators in the `pre_mutate` method.

        Parameters
        ----------
        deco_name : str
            Decorator specification of the decorator to remove. If nothing else is
            specified, all decorators matching that name will be removed.
        deco_args : List[Any], optional, default None
            Positional arguments to match the decorator specification.
        deco_kwargs : Dict[str, Any], optional, default None
            Keyword arguments to match the decorator specification.

        Returns
        -------
        bool
            Returns True if a decorator was removed.
        """

        if not self._pre_mutate:
            raise MetaflowException(
                "Removing flow-decorator '%s' from %s is only allowed in the `pre_mutate` "
                "method and not the `mutate` method" % (deco_name, self._inserted_by)
            )

        do_all = deco_args is None and deco_kwargs is None
        did_remove = False
        if do_all and deco_name in self._flow_cls._flow_decorators:
            del self._flow_cls._flow_decorators[deco_name]
            return True
        old_deco_list = self._flow_cls._flow_decorators.get(deco_name)
        if not old_deco_list:
            debug.userconf_exec(
                "Mutable flow failed to remove decorator '%s' from flow (non present)"
                % deco_name
            )
            return False
        new_deco_list = []
        for deco in old_deco_list:
            if deco.get_args_kwargs() == (deco_args or [], deco_kwargs or {}):
                did_remove = True
            else:
                new_deco_list.append(deco)
        debug.userconf_exec(
            "Mutable flow removed %d decorators from flow"
            % (len(old_deco_list) - len(new_deco_list))
        )
        if new_deco_list:
            self._flow_cls._flow_decorators[deconame] = new_deco_list
        else:
            del self._flow_cls._flow_decorators[deco_name]
        return did_remove

    def __getattr__(self, name):
        # We allow direct access to the steps, configs and parameters but nothing else
        from metaflow.parameters import Parameter

        from .mutable_step import MutableStep

        attr = getattr(self._flow_cls, name)
        if attr:
            # Steps
            if callable(attr) and hasattr(attr, "is_step"):
                return MutableStep(
                    self._flow_cls,
                    attr,
                    pre_mutate=self._pre_mutate,
                    statically_defined=self._statically_defined,
                    inserted_by=self._inserted_by,
                )
            if name[0] == "_" or name in self._flow_cls._NON_PARAMETERS:
                raise AttributeError(self, name)
            if isinstance(attr, (Parameter, ConfigValue)):
                return attr
        raise AttributeError(self, name)



import glob
import shutil
import subprocess
import re

from functools import partial
from itertools import chain
from pathlib import Path

WHITELIST = {
    "README.txt",
    "__init__.py",
    "vendor_any.txt",
    "vendor_v3_6.txt",
    "vendor_v3_7.txt",
    "pip.LICENSE",
}

# Borrowed from https://github.com/pypa/pip/tree/main/src/pip/_vendor

VENDOR_SUBDIR = re.compile(r"^_vendor/vendor_([a-zA-Z0-9_]+).txt$")


def delete_all(*paths, whitelist=frozenset()):
    for item in paths:
        if item.is_dir():
            shutil.rmtree(item, ignore_errors=True)
        elif item.is_file() and item.name not in whitelist:
            item.unlink()


def iter_subtree(path):
    """Recursively yield all files in a subtree, depth-first"""
    if not path.is_dir():
        if path.is_file():
            yield path
        return
    for item in path.iterdir():
        if item.is_dir():
            yield from iter_subtree(item)
        elif item.is_file():
            yield item


def patch_vendor_imports(file, replacements):
    text = file.read_text("utf8")
    for replacement in replacements:
        text = replacement(text)
    file.write_text(text, "utf8")


def find_vendored_libs(vendor_dir, whitelist, whitelist_dirs):
    vendored_libs = []
    paths = []
    for item in vendor_dir.iterdir():
        if item.is_dir() and item not in whitelist_dirs:
            vendored_libs.append(item.name)
        elif item.is_file() and item.name not in whitelist:
            vendored_libs.append(item.stem)  # without extension
        else:  # not a dir or a file not in the whitelist
            continue
        paths.append(item)
    return vendored_libs, paths


def fetch_licenses(*info_dir, vendor_dir):
    for file in chain.from_iterable(map(iter_subtree, info_dir)):
        if "LICENSE" in file.name:
            library = file.parent.name.split("-")[0]
            shutil.copy(file, vendor_dir / ("%s.LICENSE" % library))
        else:
            continue


def vendor(vendor_dir):
    # remove everything
    delete_all(*vendor_dir.iterdir(), whitelist=WHITELIST)

    exclude_subdirs = []
    # Iterate on the vendor*.txt files
    for vendor_file in glob.glob(f"{vendor_dir.name}/vendor*.txt"):
        # We extract the subdirectory we are going to extract into
        subdir = VENDOR_SUBDIR.match(vendor_file).group(1)
        # Includes "any" but it doesn't really matter unless you install "any"
        exclude_subdirs.append(subdir)

    for subdir in exclude_subdirs:
        create_init_file = False
        if subdir == "any":
            vendor_subdir = vendor_dir
            # target package is <parent>.<vendor_dir>; foo/_vendor -> foo._vendor
            pkgname = f"{vendor_dir.parent.name}.{vendor_dir.name}"
        else:
            create_init_file = True
            vendor_subdir = vendor_dir / subdir
            # target package is <parent>.<vendor_dir>; foo/_vendor -> foo._vendor
            pkgname = f"{vendor_dir.parent.name}.{vendor_dir.name}.{vendor_subdir.name}"

        # install with pip
        subprocess.run(
            [
                "python3",
                "-m",
                "pip",
                "install",
                "-t",
                str(vendor_subdir),
                "-r",
                "_vendor/vendor_%s.txt" % subdir,
                "--no-compile",
            ]
        )

        # fetch licenses
        fetch_licenses(*vendor_subdir.glob("*.dist-info"), vendor_dir=vendor_subdir)

        # delete stuff that's not needed
        delete_all(
            *vendor_subdir.glob("*.dist-info"),
            *vendor_subdir.glob("*.egg-info"),
            vendor_subdir / "bin",
        )

        # Touch a __init__.py file
        if create_init_file:
            with open(
                "%s/__init__.py" % str(vendor_subdir), "w+", encoding="utf-8"
            ) as f:
                f.write("# Empty file")

        vendored_libs, paths = find_vendored_libs(
            vendor_subdir, WHITELIST, exclude_subdirs
        )

        replacements = []
        for lib in vendored_libs:
            replacements += (
                partial(  # import bar -> import foo._vendor.bar
                    re.compile(r"(^\s*)import {}\n".format(lib), flags=re.M).sub,
                    r"\1from {} import {}\n".format(pkgname, lib),
                ),
                partial(  # from bar -> from foo._vendor.bar
                    re.compile(r"(^\s*)from {}(\.|\s+)".format(lib), flags=re.M).sub,
                    r"\1from {}.{}\2".format(pkgname, lib),
                ),
            )

        for file in chain.from_iterable(map(iter_subtree, paths)):
            if file.suffix == ".py":
                patch_vendor_imports(file, replacements)


if __name__ == "__main__":
    here = Path("__file__").resolve().parent
    vendor_tl_dir = here / "_vendor"
    has_vendor_file = len(glob.glob(f"{vendor_tl_dir.name}/vendor*.txt")) > 0
    assert has_vendor_file, "_vendor/vendor*.txt file not found"
    assert (
        vendor_tl_dir / "__init__.py"
    ).exists(), "_vendor/__init__.py file not found"
    vendor(vendor_tl_dir)



import os
import shutil
import sys
import tempfile
import zlib
import base64
from functools import wraps
from io import BytesIO
from itertools import takewhile
import re


try:
    # python2
    unicode_type = unicode
    bytes_type = str
    from urllib import quote, unquote

    # unquote_bytes should be a function that takes a urlencoded byte
    # string, encoded in UTF-8, url-decodes it and returns it as a
    # unicode object. Confusingly, how to accomplish this differs
    # between Python2 and Python3.
    #
    # Test with this input URL:
    # b'crazypath/%01%C3%B'
    # it should produce
    # u'crazypath/\x01\xff'
    def unquote_bytes(x):
        return to_unicode(unquote(to_bytes(x)))

    # this is used e.g. by mflog/save_logs.py to identify paths
    class Path(object):
        def __init__(self, path):
            self.path = path

        def __str__(self):
            return self.path

    from pipes import quote as _quote
except NameError:
    # python3
    unicode_type = str
    bytes_type = bytes
    from urllib.parse import quote, unquote
    from pathlib import Path

    def unquote_bytes(x):
        return unquote(to_unicode(x))

    from shlex import quote as _quote


class TempDir(object):
    # Provide a temporary directory since Python 2.7 does not have it inbuilt
    def __enter__(self):
        self.name = tempfile.mkdtemp()
        return self.name

    def __exit__(self, exc_type, exc_value, traceback):
        shutil.rmtree(self.name)


def cached_property(getter):
    @wraps(getter)
    def exec_once(self):
        saved_name = "__%s" % getter.__name__
        if not hasattr(self, saved_name):
            setattr(self, saved_name, getter(self))
        return getattr(self, saved_name)

    return property(exec_once)


def all_equal(it):
    """
    Return True if all elements of the given iterator are equal.
    """
    it = iter(it)
    try:
        first = next(it)
    except StopIteration:
        return True
    for x in it:
        if x != first:
            return False
    return True


def url_quote(url):
    """
    Encode a unicode URL to a safe byte string
    """
    # quote() works reliably only with (byte)strings in Python2,
    # hence we need to .encode('utf-8') first. To see by yourself,
    # try quote(u'\xff') in python2. Python3 converts the output
    # always to Unicode, hence we need the outer to_bytes() too.
    #
    # We mark colon as a safe character to keep simple ASCII urls
    # nice looking, e.g. "http://google.com"
    return to_bytes(quote(to_bytes(url), safe="/:"))


def url_unquote(url_bytes):
    """
    Decode a byte string encoded with url_quote to a unicode URL
    """
    return unquote_bytes(url_bytes)


def is_stringish(x):
    """
    Returns true if the object is a unicode or a bytes object
    """
    return isinstance(x, bytes_type) or isinstance(x, unicode_type)


def to_fileobj(x):
    """
    Convert any string-line object to a byte-returning fileobj
    """
    return BytesIO(to_bytes(x))


def to_unicode(x):
    """
    Convert any object to a unicode object
    """
    if isinstance(x, bytes_type):
        return x.decode("utf-8")
    else:
        return unicode_type(x)


def to_bytes(x):
    """
    Convert any object to a byte string
    """
    if isinstance(x, unicode_type):
        return x.encode("utf-8")
    elif isinstance(x, bytes_type):
        return x
    elif isinstance(x, float):
        return repr(x).encode("utf-8")
    else:
        return str(x).encode("utf-8")


def get_username():
    """
    Return the name of the current user, or None if the current user
    could not be determined.
    """
    from metaflow.metaflow_config import USER

    # note: the order of the list matters
    ENVVARS = ["SUDO_USER", "USERNAME", "USER"]
    for user in [USER] + [os.environ.get(x) for x in ENVVARS]:
        if user and user != "root":
            return user
    return None


def resolve_identity_as_tuple():
    from metaflow.exception import MetaflowUnknownUser

    prod_token = os.environ.get("METAFLOW_PRODUCTION_TOKEN")
    if prod_token:
        return "production", prod_token
    user = get_username()
    if user and user != "root":
        return "user", user
    else:
        raise MetaflowUnknownUser()


def resolve_identity():
    identity_type, identity_value = resolve_identity_as_tuple()
    return "%s:%s" % (identity_type, identity_value)


def get_latest_run_id(echo, flow_name):
    from metaflow.plugins.datastores.local_storage import LocalStorage

    local_root = LocalStorage.datastore_root
    if local_root is None:
        local_root = LocalStorage.get_datastore_root_from_config(
            echo, create_on_absent=False
        )
    if local_root:
        path = os.path.join(local_root, flow_name, "latest_run")
        if os.path.exists(path):
            with open(path) as f:
                return f.read()
    return None


def write_latest_run_id(obj, run_id):
    from metaflow.plugins.datastores.local_storage import LocalStorage

    if LocalStorage.datastore_root is None:
        LocalStorage.datastore_root = LocalStorage.get_datastore_root_from_config(
            obj.echo
        )
    path = LocalStorage.path_join(LocalStorage.datastore_root, obj.flow.name)
    try:
        os.makedirs(path)
    except OSError as x:
        if x.errno != 17:
            # Directories exists in other case which is fine
            raise
    with open(os.path.join(path, "latest_run"), "w") as f:
        f.write(str(run_id))


def get_object_package_version(obj):
    """
    Return the top level package name and package version that defines the
    class of the given object.
    """
    try:
        module_name = obj.__class__.__module__

        if "." in module_name:
            top_package_name = module_name.split(".")[0]
        else:
            top_package_name = module_name

    except AttributeError:
        return None, None

    try:
        top_package_version = sys.modules[top_package_name].__version__
        return top_package_name, top_package_version

    except AttributeError:
        return top_package_name, None


def compress_list(lst, separator=",", rangedelim=":", zlibmarker="!", zlibmin=500):
    from metaflow.exception import MetaflowInternalError

    bad_items = [x for x in lst if separator in x or rangedelim in x or zlibmarker in x]
    if bad_items:
        raise MetaflowInternalError(
            "Item '%s' includes a delimiter character "
            "so it can't be compressed" % bad_items[0]
        )
    # Three output modes:
    lcp = longest_common_prefix(lst)
    if len(lst) < 2 or not lcp:
        # 1. Just a comma-separated list
        res = separator.join(lst)
    else:
        # 2. Prefix and a comma-separated list of suffixes
        lcplen = len(lcp)
        residuals = [e[lcplen:] for e in lst]
        res = rangedelim.join((lcp, separator.join(residuals)))
    if len(res) < zlibmin:
        return res
    else:
        # 3. zlib-compressed, base64-encoded, prefix-encoded list

        # interestingly, a typical zlib-encoded list of suffixes
        # has plenty of redundancy. Decoding the data *twice* helps a
        # lot
        compressed = zlib.compress(zlib.compress(to_bytes(res)))
        return zlibmarker + base64.b64encode(compressed).decode("utf-8")


def decompress_list(lststr, separator=",", rangedelim=":", zlibmarker="!"):
    # Three input modes:
    if lststr[0] == zlibmarker:
        # 3. zlib-compressed, base64-encoded
        lstbytes = base64.b64decode(lststr[1:])
        decoded = zlib.decompress(zlib.decompress(lstbytes)).decode("utf-8")
    else:
        decoded = lststr

    if rangedelim in decoded:
        prefix, suffixes = decoded.split(rangedelim)
        # 2. Prefix and a comma-separated list of suffixes
        return [prefix + suffix for suffix in suffixes.split(separator)]
    else:
        # 1. Just a comma-separated list
        return decoded.split(separator)


def longest_common_prefix(lst):
    if lst:
        return "".join(
            a for a, _ in takewhile(lambda t: t[0] == t[1], zip(min(lst), max(lst)))
        )
    else:
        return ""


def get_metaflow_root():
    return os.path.dirname(os.path.dirname(__file__))


def dict_to_cli_options(params):
    # Prevent circular imports
    from .user_configs.config_options import ConfigInput

    for k, v in params.items():
        # Omit boolean options set to false or None, but preserve options with an empty
        # string argument.
        if v is not False and v is not None:
            # we need special handling for 'with' since it is a reserved
            # keyword in Python, so we call it 'decospecs' in click args
            if k == "decospecs":
                k = "with"
            if k in ("config", "config_value"):
                # Special handling here since we gather them all in one option but actually
                # need to send them one at a time using --config-value <name> kv.<name>
                # Note it can be either config or config_value depending
                # on click processing order.
                for config_name in v.keys():
                    yield "--config-value"
                    yield to_unicode(config_name)
                    yield to_unicode(ConfigInput.make_key_name(config_name))
                continue
            if k == "local_config_file":
                # Skip this value -- it should only be used locally and never when
                # forming another command line
                continue
            k = k.replace("_", "-")
            v = v if isinstance(v, (list, tuple, set)) else [v]
            for value in v:
                yield "--%s" % k
                if not isinstance(value, bool):
                    value = to_unicode(value)

                    # Of the value starts with $, assume the caller wants shell variable
                    # expansion to happen, so we pass it as is.
                    # NOTE: We strip '\' to allow for various storages to use escaped
                    # shell variables as well.
                    if value.lstrip("\\").startswith("$"):
                        yield value
                    else:
                        # Otherwise, assume it is a literal value and quote it safely
                        yield _quote(value)


# This function is imported from https://github.com/cookiecutter/whichcraft
def which(cmd, mode=os.F_OK | os.X_OK, path=None):
    """Given a command, mode, and a PATH string, return the path which
    conforms to the given mode on the PATH, or None if there is no such
    file.
    `mode` defaults to os.F_OK | os.X_OK. `path` defaults to the result
    of os.environ.get("PATH"), or can be overridden with a custom search
    path.
    Note: This function was backported from the Python 3 source code.
    """
    # Check that a given file can be accessed with the correct mode.
    # Additionally check that `file` is not a directory, as on Windows
    # directories pass the os.access check.
    try:  # Forced testing
        from shutil import which as w

        return w(cmd, mode, path)
    except ImportError:

        def _access_check(fn, mode):
            return os.path.exists(fn) and os.access(fn, mode) and not os.path.isdir(fn)

        # If we're given a path with a directory part, look it up directly
        # rather than referring to PATH directories. This includes checking
        # relative to the current directory, e.g. ./script
        if os.path.dirname(cmd):
            if _access_check(cmd, mode):
                return cmd
            return None

        if path is None:
            path = os.environ.get("PATH", os.defpath)
        if not path:
            return None

        path = path.split(os.pathsep)

        files = [cmd]
        seen = set()
        for dir in path:
            normdir = os.path.normcase(dir)
            if normdir not in seen:
                seen.add(normdir)
                for thefile in files:
                    name = os.path.join(dir, thefile)
                    if _access_check(name, mode):
                        return name

        return None


def to_camelcase(obj):
    """
    Convert all keys of a json to camel case from snake case.
    """
    if isinstance(obj, (str, int, float)):
        return obj
    if isinstance(obj, dict):
        res = obj.__class__()
        for k in obj:
            res[re.sub(r"(?!^)_([a-zA-Z])", lambda x: x.group(1).upper(), k)] = (
                to_camelcase(obj[k])
            )
    elif isinstance(obj, (list, set, tuple)):
        res = obj.__class__(to_camelcase(v) for v in obj)
    else:
        return obj
    return res


def to_pascalcase(obj):
    """
    Convert all keys of a json to pascal case.
    """
    if isinstance(obj, (str, int, float)):
        return obj
    if isinstance(obj, dict):
        res = obj.__class__()
        for k in obj:
            res[re.sub("([a-zA-Z])", lambda x: x.groups()[0].upper(), k, count=1)] = (
                to_pascalcase(obj[k])
            )
    elif isinstance(obj, (list, set, tuple)):
        res = obj.__class__(to_pascalcase(v) for v in obj)
    else:
        return obj
    return res


def tar_safe_extract(tar, path=".", members=None, *, numeric_owner=False):
    def is_within_directory(abs_directory, target):
        prefix = os.path.commonprefix([abs_directory, os.path.abspath(target)])
        return prefix == abs_directory

    abs_directory = os.path.abspath(path)
    if any(
        not is_within_directory(abs_directory, os.path.join(path, member.name))
        for member in tar.getmembers()
    ):
        raise Exception("Attempted path traversal in TAR file")

    tar.extractall(path, members, numeric_owner=numeric_owner)


def to_pod(value):
    """
    Convert a python object to plain-old-data (POD) format.

    Parameters
    ----------
    value : Any
        Value to convert to POD format. The value can be a string, number, list,
        dictionary, or a nested structure of these types.
    """
    # Prevent circular imports
    from metaflow.parameters import DeployTimeField

    if isinstance(value, (str, int, float)):
        return value
    if isinstance(value, dict):
        return {to_pod(k): to_pod(v) for k, v in value.items()}
    if isinstance(value, (list, set, tuple)):
        return [to_pod(v) for v in value]
    if isinstance(value, DeployTimeField):
        return value.print_representation
    return str(value)


from metaflow._vendor.packaging.version import parse as version_parse



import sys
import subprocess
from pathlib import Path
import sysconfig
import site


def find_makefile():
    possible_dirs = []

    # 1) The standard sysconfig-based location
    data_dir = sysconfig.get_paths()["data"]
    possible_dirs.append(Path(data_dir) / "share" / "metaflow" / "devtools")

    # 2) The user base (e.g. ~/.local on many systems)
    user_base = site.getuserbase()  # e.g. /home/runner/.local
    possible_dirs.append(Path(user_base) / "share" / "metaflow" / "devtools")

    # 3) site-packages can vary, we can guess share/.. near each site-packages
    # (Works if pip actually placed devtools near site-packages.)
    for p in site.getsitepackages():
        possible_dirs.append(Path(p).parent / "share" / "metaflow" / "devtools")
    user_site = site.getusersitepackages()
    possible_dirs.append(Path(user_site).parent / "share" / "metaflow" / "devtools")

    for candidate_dir in possible_dirs:
        makefile_candidate = candidate_dir / "Makefile"
        if makefile_candidate.is_file():
            return makefile_candidate

    return None


def main():
    makefile_path = find_makefile()
    if not makefile_path:
        print("ERROR: Could not find executable in any known location.")
        sys.exit(1)
    cmd = ["make", "-f", str(makefile_path)] + sys.argv[1:]

    try:
        completed = subprocess.run(cmd, check=True)
        sys.exit(completed.returncode)
    except subprocess.CalledProcessError as ex:
        sys.exit(ex.returncode)
    except KeyboardInterrupt:
        print("Process interrupted by user. Exiting cleanly.")
        sys.exit(1)
